Attribute VB_Name = "vars"
Option Explicit

'_________________________________________________________
'Dieses Modul enthlt alle ffentlichen Variablen
'und Funktionen zur Suche nach Datenstzen in der
'Kanten, Knoten etc. Datenbank
'und mathematische Funktionen
'
'SetEdge(...) Kante erstellen
'RemoveEdge(...) Kante entfernen
'SetNode(...) Knoten erstellen
'SetObst() Hindernis erstellen
'RemoveObstacle() Hindernis lschen
'
'Die beiden folgenden Prozeduren machen die Wegfindung
'RemoveCutEdges()   Hindernisse mit Kanten schneiden
'SetGraph() Knoten um ein Hindernis herum setzen
'
'ShortestPath() Dijkstras Algorithmus
'_________________________________________________________

'Wert, der grer ist als die grte mgliche Strecke
Public Const cINFINITY As Long = 1000000000

Public Const cPI As Double = 3.14159265358979

Public Type tNodes                              'Knoten
    XPos As Long                                'an den Koordinaten
    YPos As Long                                'XPos/YPos
    Statelifted As Boolean                      'Knotenstatus
    ConnectedToNode() As Boolean                'Verbunden mit Knoten im Index?
    ConnectedOverEdge() As Long                 'Verbunden mit Knoten im
End Type                                        'Index ber Kante

Public Type tEdges
    StartNodeIndex As Long                      'Kante geht von hier
    EndNodeIndex As Long                        'bis hier
    length As Double                            'Lnge der Kante
End Type

Public Type tObstacles                          'Hindernisse
    X1Pos As Double                             'X-Position des 1. Punkts
    Y1Pos As Double                             'Y-Position des 1. Punkts
    X2Pos As Double                             'X-Position des 2. Punkts
    Y2Pos As Double                             'Y-Position des 2. Punkts
End Type

Public Nodes() As tNodes                        'Array fr Knoten
Public Edges() As tEdges                        'Array fr Kanten

Public Obstacles() As tObstacles                'Array fr Hindernisse

'Lnge der Krzesten Verbindung zum Knoten mit dem selben Index
Public dDistance() As Double
                                                
'lToNode(v) = u; d.h. man muss ber u, um zu v zu kommmen
Public lToNode() As Long

Public lWaitingNodes As Long                    'Anzahl der Knoten auf der Warteliste
Public lWaitingList() As Long                   'Warteliste

Public lCountNodes As Long                      'Anzahl der Knoten
Public lCountEdges As Long                      '...        Kanten

Public lCountObsts As Long                      'Anzahl der Hindernisse

Public dCarSpeed As Double                      'Bewegungsgeschwindigkeit des Autos

'Variablen frs Abfahren
Public lStartNodeIndex As Long                  'Startknoten fr Berechnung und spter Bewegung
Public lLastNodeIndex As Long                   'letzter angefahrener Knoten
Public lNextNodeIndex As Long                   'nchster anzufahrender Knoten
Public lDestNodeIndex As Long                   'Zielknoten
Public lCurrentEdgeIndex As Long                'aktuelle Kante
Public dgegkat As Double, dankat As Double      'fr Berechnung des Bewegungswinkels
Public dDirection As Double                     'Bewegungsrichtung ber die aktuelle Kante
Public dCarPosX As Double                       'X-Koordinate des Autos
Public dCarPosY As Double                       'Y-Koordinate des Autos

Public lTimeCounter As Long                     'Zhler fr Rechendauer

Public bShortestWaysReady As Boolean            'Ist der Weg berechnet? Noch aktuell?

Public lActMousePosX As Long, lActMousePosY As Long   'Mausposition

Public lMoveIndex As Long                       'Index fr Objekt, das bewegt wird
Public lMoveObstStartNode As Boolean            'Wird das erste Ende bewegt?

Public lDragStartX As Long                      'Startkoordinaten frs Verwschieben von Objekten
Public lDragStartY As Long
Public bDragState As Boolean                    'wird gerade etwas gezogen?

Public bCreateObst As Boolean                   'Wird gerade ein Hindernis gemacht?

Public sBGPicPath As String                     'Pfad zum Hintergrundbild

Public Function FindNextNode(lFromX As Long, lFromY As Long)
    'nchsten Knoten von einer Position, hier meist Mausposition mit Pythagoras ermitteln
    
    Dim lLowestDistance As Long, lDistance As Double, n As Long
    
    On Error GoTo seterrorcode
    
    lLowestDistance = -1
    
    If lCountNodes = 0 Then GoTo seterrorcode
    For n = 0 To lCountNodes - 1
        lDistance = PythA(lFromX - Nodes(n).XPos, lFromY - Nodes(n).YPos)
        If lLowestDistance = -1 Or lLowestDistance > lDistance Then
            lLowestDistance = lDistance
            FindNextNode = n
        End If
    Next n
    
seterrorcode:
    If lLowestDistance = -1 Then FindNextNode = -1
    
End Function

Public Function PythA(ByVal a As Single, ByVal b As Single) As Single   'Pythagoras addition

    PythA = Sqr(a * a + b * b)

End Function

Public Sub ResetGraph()
    
    Dim n As Long
    
    ReDim Preserve Nodes(1)              'initialisieren
    ReDim Nodes(0).ConnectedToNode(1)
    ReDim Nodes(1).ConnectedToNode(1)
    ReDim Nodes(0).ConnectedOverEdge(1)
    ReDim Nodes(1).ConnectedOverEdge(1)
    Erase Edges
    ReDim dDistance(1)
    ReDim lToNode(1)
    
    ReDim lWaitingList(0)
    lWaitingNodes = 0
    
    lCountNodes = 2
    lCountEdges = 0
        
    lLastNodeIndex = -1
    lNextNodeIndex = -1
    lStartNodeIndex = 0
    lDestNodeIndex = 1
    bShortestWaysReady = False
        
    frmpathfind.picbg.Cls                           'alte Anzeige lschen
    
    frmpathfind.DrawCursorCross                     'Fadenkreuz
        
    frmpathfind.DrawEdges                           'Kanten zeichnen
    
    frmpathfind.DrawNodes                           'Knoten zeichnen
    
    frmpathfind.DrawObstacles                       'Hindernisse zeichnen
    
End Sub

Public Sub RemoveCutEdges()
    
    Dim n As Long, m As Long                    'Zhlvariablen
    Dim dXDiff As Double, dYDiff As Double      'Unterschiede zwischen Start- und Endnoten
    Dim dAscentEdge As Double                   'Steigung der Kante (als Gerade betrachten)
    Dim dAscentObst As Double                   'Steigung der Kante (als Gerade betrachten)
    Dim dCutX As Double, dCutY As Double        'Koordinaten des Schnittpunkts
    Dim dAbszissEdge As Double                  'Abszisse der Kante
    Dim dAbszissObst As Double                  'Abszisse des Hindernisses
    Dim bCutEdgesObst As Boolean                'schneidet die Kante ein Hindernis?
    
    Do While n < lCountEdges                    'alle Kanten mit
        bCutEdgesObst = False
        
        dXDiff = Nodes(Edges(n).StartNodeIndex).XPos - Nodes(Edges(n).EndNodeIndex).XPos   'Steigung der Kante
        dYDiff = Nodes(Edges(n).StartNodeIndex).YPos - Nodes(Edges(n).EndNodeIndex).YPos
        If dXDiff = 0 Then dXDiff = 0.000000000000001                      'Division durch Null verhindern
        dAscentEdge = dYDiff / dXDiff                                      'geringe Ungenauigkeit macht nichts
        
        dAbszissEdge = Nodes(Edges(n).StartNodeIndex).YPos - (Nodes(Edges(n).StartNodeIndex).XPos * dAscentEdge)   'Abszissen
        
        Do While m < lCountObsts                'allen Hindernissen schneiden
            dXDiff = Obstacles(m).X1Pos - Obstacles(m).X2Pos      'Steigung des Hindernisses
            dYDiff = Obstacles(m).Y1Pos - Obstacles(m).Y2Pos
            If dXDiff = 0 Then dXDiff = 0.000000000000001
            dAscentObst = dYDiff / dXDiff
            
            dAbszissObst = Obstacles(m).Y1Pos - (Obstacles(m).X1Pos * dAscentObst)                  'berechnen
            
            If dAscentEdge <> dAscentObst Then                                      'Parallelen ignorieren
                dCutX = Round((dAbszissObst - dAbszissEdge) / (dAscentEdge - dAscentObst))  'Schnittpunkt X
                dCutY = Round(dAscentEdge * dCutX + dAbszissEdge)                           'Y - Wert
                
                If Nodes(Edges(n).StartNodeIndex).XPos < Nodes(Edges(n).EndNodeIndex).XPos Then     'Schnittpunkt muss auf der Strecke liegen
                    If dCutX > Nodes(Edges(n).StartNodeIndex).XPos And dCutX < Nodes(Edges(n).EndNodeIndex).XPos Then '!
                        If Obstacles(m).X1Pos < Obstacles(m).X2Pos Then
                            If dCutX > Obstacles(m).X1Pos And dCutX < Obstacles(m).X2Pos Then
                                bCutEdgesObst = True
                                Exit Do
                           End If
                        Else
                            If dCutX > Obstacles(m).X2Pos And dCutX < Obstacles(m).X1Pos Then
                                bCutEdgesObst = True
                                Exit Do
                            End If
                        End If
                    End If
                Else
                    If dCutX > Nodes(Edges(n).EndNodeIndex).XPos And dCutX < Nodes(Edges(n).StartNodeIndex).XPos Then
                        If Obstacles(m).X1Pos < Obstacles(m).X2Pos Then
                            If dCutX > Obstacles(m).X1Pos And dCutX < Obstacles(m).X2Pos Then
                                bCutEdgesObst = True
                                Exit Do
                            End If
                        Else
                            If dCutX > Obstacles(m).X2Pos And dCutX < Obstacles(m).X1Pos Then
                                bCutEdgesObst = True
                                Exit Do
                           End If
                        End If
                    End If
                End If
            End If
            DoEvents
            m = m + 1
        Loop
        If bCutEdgesObst Then
            RemoveEdge n
            n = n - 1
        End If
        m = 0
        n = n + 1
    Loop
    
End Sub

Public Sub SetGraph()
    
    Dim n As Long, m As Long
    
    For n = 0 To lCountObsts - 1
        If Obstacles(n).X1Pos > Obstacles(n).X2Pos Then 'Knoten etwas von den Hindernissen entfernen
            If Obstacles(n).Y1Pos > Obstacles(n).Y2Pos Then
                SetNode Obstacles(n).X1Pos, Obstacles(n).Y1Pos
                SetNode Obstacles(n).X2Pos, Obstacles(n).Y2Pos
            Else
                SetNode Obstacles(n).X1Pos, Obstacles(n).Y1Pos
                SetNode Obstacles(n).X2Pos, Obstacles(n).Y2Pos
            End If
        Else
            If Obstacles(n).Y1Pos > Obstacles(n).Y2Pos Then
                SetNode Obstacles(n).X1Pos, Obstacles(n).Y1Pos
                SetNode Obstacles(n).X2Pos, Obstacles(n).Y2Pos
            Else
                SetNode Obstacles(n).X1Pos, Obstacles(n).Y1Pos
                SetNode Obstacles(n).X2Pos, Obstacles(n).Y2Pos
            End If
        End If
    Next n
    
    For n = 0 To lCountNodes - 1    'Neuen Knoten mit anderen verbinden
        For m = 0 To lCountNodes - 1
            SetEdge n, m
        Next m
    Next n
    
End Sub

Public Sub SetObst(ByVal X1Pos As Double, ByVal Y1Pos As Double, ByVal X2Pos As Double, ByVal Y2Pos As Double)
    
    lCountObsts = lCountObsts + 1
    ReDim Preserve Obstacles(lCountObsts - 1)
    Obstacles(lCountObsts - 1).X1Pos = X1Pos
    Obstacles(lCountObsts - 1).Y1Pos = Y1Pos
    Obstacles(lCountObsts - 1).X2Pos = X2Pos
    Obstacles(lCountObsts - 1).Y2Pos = Y2Pos
    
End Sub

Public Sub RemoveObstacle(Index As Long)
    
    Dim n As Long, m As Long
    
    For n = Index To lCountObsts - 2                    'nachrcken
        Obstacles(n).X1Pos = Obstacles(n + 1).X1Pos
        Obstacles(n).X2Pos = Obstacles(n + 1).X2Pos
        Obstacles(n).Y1Pos = Obstacles(n + 1).Y1Pos
        Obstacles(n).Y2Pos = Obstacles(n + 1).Y2Pos
    Next n
        
    lCountObsts = lCountObsts - 1                       'redimensionieren
    If lCountObsts > 0 Then ReDim Preserve Obstacles(0 To lCountObsts - 1)
    
End Sub

Public Sub SetEdge(lBeginningNodeIndex As Long, lEndNodeIndex As Long)
    
    Dim lSelectedNodeIndex As Long, n As Long
    
    If lBeginningNodeIndex = lEndNodeIndex Then Exit Sub                    'Nicht mit sich selbst verbinden
    If Nodes(lBeginningNodeIndex).ConnectedToNode(lEndNodeIndex) Then Exit Sub   'Wenn schon verbunden, dann abbrechen
    
    lCountEdges = lCountEdges + 1                                       'Kantenzhler erhhen
    ReDim Preserve Edges(0 To lCountEdges - 1)
    Edges(lCountEdges - 1).StartNodeIndex = lBeginningNodeIndex         'Startknoten setzen
    Edges(lCountEdges - 1).EndNodeIndex = lEndNodeIndex                 'Endknoten setzen
        
    'Lnge der neuen Kante berechnen
    Edges(lCountEdges - 1).length = PythA(Nodes(Edges(lCountEdges - 1).StartNodeIndex).XPos - Nodes(Edges(lCountEdges - 1).EndNodeIndex).XPos, _
                                          Nodes(Edges(lCountEdges - 1).StartNodeIndex).YPos - Nodes(Edges(lCountEdges - 1).EndNodeIndex).YPos)
    
    bShortestWaysReady = False
            
    'Verbindungsinformationen: welcher Knoten ist mit welchem verbunden?
    '                          und ber welche Kante?
    Nodes(Edges(lCountEdges - 1).StartNodeIndex).ConnectedToNode(Edges(lCountEdges - 1).EndNodeIndex) = True
    Nodes(Edges(lCountEdges - 1).EndNodeIndex).ConnectedToNode(Edges(lCountEdges - 1).StartNodeIndex) = True
    Nodes(Edges(lCountEdges - 1).StartNodeIndex).ConnectedOverEdge(Edges(lCountEdges - 1).EndNodeIndex) = lCountEdges - 1
    Nodes(Edges(lCountEdges - 1).EndNodeIndex).ConnectedOverEdge(Edges(lCountEdges - 1).StartNodeIndex) = lCountEdges - 1

End Sub

Public Sub RemoveEdge(lEdgeIndex As Long)
    
    Dim n As Long, m As Long
    
    'Kanten nachrcken
    For n = lEdgeIndex To lCountEdges - 2
        Edges(n).EndNodeIndex = Edges(n + 1).EndNodeIndex
        Edges(n).StartNodeIndex = Edges(n + 1).StartNodeIndex
        Edges(n).length = Edges(n + 1).length
    Next n
        
    lCountEdges = lCountEdges - 1
    If lCountEdges > 0 Then ReDim Preserve Edges(0 To lCountEdges - 1)
    
    For n = 0 To lCountNodes - 1
        For m = 0 To lCountNodes - 1
            If Nodes(n).ConnectedOverEdge(m) > lEdgeIndex Then
                Nodes(n).ConnectedOverEdge(m) = Nodes(n).ConnectedOverEdge(m) - 1
            ElseIf Nodes(n).ConnectedOverEdge(m) = lEdgeIndex Then
                Nodes(n).ConnectedToNode(m) = False
            End If
        Next m
    Next n

End Sub

Public Sub SetNode(XPos As Double, YPos As Double)
    
    Dim n As Long
    
    For n = 0 To lCountNodes - 1          'neuer Knoten muss min. 10 Pixel von allen anderen Knoten entfernt sein
        If PythA(Nodes(n).XPos - lActMousePosX, _
            Nodes(n).YPos - lActMousePosY) < 150 Then Exit Sub
    Next n
    
    lCountNodes = lCountNodes + 1                                       'Zhler erhhen
    ReDim Preserve Nodes(0 To lCountNodes - 1)
    ReDim Preserve lToNode(0 To lCountNodes - 1)
    ReDim Preserve dDistance(0 To lCountNodes - 1)
    
    Nodes(lCountNodes - 1).XPos = XPos                                  'Position speichern
    Nodes(lCountNodes - 1).YPos = YPos
    
    For n = 0 To lCountNodes - 1                                        'alle dDistance(n) = unendlich
        dDistance(n) = cINFINITY
        ReDim Preserve Nodes(n).ConnectedToNode(0 To lCountNodes - 1)   'Verbindungsinformationen erweitern
        ReDim Preserve Nodes(n).ConnectedOverEdge(0 To lCountNodes - 1)
    Next n
    
    bShortestWaysReady = False
    
End Sub

Public Function GetNextOnList() As Long
    'nchsten Knoten auf der Warteliste
    
    Dim n As Long, dShortestLen As Double
    
    dShortestLen = cINFINITY
    GetNextOnList = -1
    For n = 0 To lWaitingNodes - 1
        If dDistance(lWaitingList(n)) < dShortestLen Then
            dShortestLen = dDistance(lWaitingList(n))
            GetNextOnList = lWaitingList(n)
        End If
    Next n
    
End Function

Public Sub AddToList(NodeIndex As Long)
    'Knoten zur Warteliste hinzufgen
    
    Dim n As Long
    
    For n = 0 To lWaitingNodes - 1
        If lWaitingList(n) = NodeIndex Then Exit Sub
    Next n
    
    lWaitingNodes = lWaitingNodes + 1
    ReDim Preserve lWaitingList(0 To lWaitingNodes - 1)
    lWaitingList(lWaitingNodes - 1) = NodeIndex
    
End Sub

Public Sub RemoveFromList(NodeIndex As Long)
    'Knoten von Warteliste entfernen
    
    Dim n As Long, m As Long
    
    If n = lWaitingNodes Then Exit Sub  'wenn NodeIndex nicht vorhanden ist
    Do Until lWaitingList(n) = NodeIndex
        n = n + 1
        If n = lWaitingNodes Then Exit Sub  'wenn NodeIndex nicht vorhanden ist
    Loop
    
    
    For m = n To lWaitingNodes - 2
        lWaitingList(m) = lWaitingList(m + 1)
    Next m
    
    lWaitingNodes = lWaitingNodes - 1
    If lWaitingNodes > 0 Then ReDim Preserve lWaitingList(0 To lWaitingNodes - 1)
    
End Sub

Public Sub ShortestPath(lFromNodeIndex As Long)
    
    Dim n As Long, u As Long, v As Long
    
    '#####################________Vorbereitung________#####################
    
    u = 0
    
    For n = 0 To lCountNodes - 1                            'alle Knoten warten
        Nodes(n).Statelifted = False
        dDistance(n) = cINFINITY                            'alle dDistance(n) = unendlich
    Next n
    
    dDistance(lFromNodeIndex) = 0                           'Distanz von Startknoten zu Startknoten = 0
    AddToList lFromNodeIndex
    v = lFromNodeIndex                  'nchsten auf der Liste
    
    lTimeCounter = 0                         'Zeitmessung beginnen
    frmpathfind.tmrcalcduration.Enabled = True
    
    '#####################________Dijkstra-Algorithmus_Anfang________#####################
    
    Do Until v = -1                     'solange wartende Knoten da sind
    
        Nodes(v).Statelifted = True     'v = Knoten mit kleinsten dDistance(v) auswhlen und anheben
        RemoveFromList v                'angehobenen Knoten von der Liste entfernen
        
        For u = 0 To lCountNodes - 1
        
            If Nodes(v).ConnectedToNode(u) Then
            'wenn v mit u verbunden ist, dann Knoten auf Warteliste und Distanz berechnen
                If dDistance(v) + Edges(Nodes(v).ConnectedOverEdge(u)).length < dDistance(u) Then
                    dDistance(u) = dDistance(v) + Edges(Nodes(v).ConnectedOverEdge(u)).length
                    lToNode(u) = v 'gehrt nicht zum Algortihmus, ist aber frs Abfahren erforderlich
                    AddToList u
                End If
            End If
            
        Next u
        DoEvents                        'wegen "mglicher" Endlosschleife bei der Entwicklung
        
        v = GetNextOnList()             'nchsten auf der Liste
        
    Loop
    
    '#####################________Dijkstra-Algorithmus_Ende________#####################
    
    frmpathfind.tmrcalcduration.Enabled = False   'Zeitmessung beenden, Ergebnis anzeigen
    If lTimeCounter = 0 Then
        frmpathfind.lblmessage.Caption = "Rechendauer: weniger als 1 Millisekunde"
    Else
        frmpathfind.lblmessage.Caption = "Rechendauer: " & lTimeCounter & " Millisekunden"
    End If
    
    bShortestWaysReady = True    'bereit fr Abfahrt
    
End Sub
